# Model for the stochastic pooling problem
# based on Li et al., AIChE Journal, 2011


# CREATE STOCHASTIC MODEL

m = StochasticModel(size(Scenarios,1));


# FIRST-STAGE MODEL

#----------------------------------------------
#----------	First-stage variables	-----------
#----------------------------------------------

@defVar(m, y_S[Sources], Bin);
@defVar(m, y_P[Pools], Bin);
@defVar(m, y_SP[source=Sources,pool=Pools; T_SP[source,pool]], Bin);
@defVar(m, y_PP[pool=Pools,pool2=Pools; T_PP[pool,pool2]], Bin);
@defVar(m, y_T[Terminals], Bin);
@defVar(m, y_ST[source=Sources,terminal=Terminals; T_ST[source,terminal]], Bin);
@defVar(m, y_PT[pool=Pools,terminal=Terminals; T_PT[pool,terminal]], Bin);

#----------------------------------------------
#----------	First-stage constraints	-----------
#----------------------------------------------

# First-stage objective
@setObjective(m, Min, sum{Investment_cost_S[source]*y_S[source], source=Sources} +
						sum{Investment_cost_P[pool]*y_P[pool], pool=Pools} +
						sum{Investment_cost_SP[source,pool]*y_SP[source,pool], source=Sources, pool=Pools; T_SP[source,pool]} +
						sum{Investment_cost_PP[pool,pool2]*y_PP[pool,pool2], pool=Pools, pool2=Pools; T_PP[pool,pool2]} +
						sum{Investment_cost_T[terminal]*y_T[terminal], terminal=Terminals} +
						sum{Investment_cost_ST[source,terminal]*y_ST[source,terminal], source=Sources, terminal=Terminals; T_ST[source,terminal]} +
						sum{Investment_cost_PT[pool,terminal]*y_PT[pool,terminal], pool=Pools, terminal=Terminals; T_PT[pool,terminal]});

# Topology constraints
@addConstraint(m, Topology_S1[source=Sources,pool=Pools; T_SP[source,pool]],
					y_S[source] >= y_SP[source,pool]);
@addConstraint(m, Topology_S2[source=Sources,terminal=Terminals; T_ST[source,terminal]],
					y_S[source] >= y_ST[source,terminal]);
@addConstraint(m, Topology_S3[source=Sources],
					sum{y_SP[source,pool], pool=Pools; T_SP[source,pool]} +
					sum{y_ST[source,terminal], terminal=Terminals; T_ST[source,terminal]} >= y_S[source]);

@addConstraint(m, Topology_P1[source=Sources,pool=Pools; T_SP[source,pool]],
					y_P[pool] >= y_SP[source,pool]);
@addConstraint(m, Topology_P2[pool=Pools,pool2=Pools; T_PP[pool,pool2]],
					y_P[pool] >= y_PP[pool,pool2]);
@addConstraint(m, Topology_P3[pool=Pools,pool2=Pools; T_PP[pool,pool2]],
					y_P[pool2] >= y_PP[pool,pool2]);
@addConstraint(m, Topology_P4[pool=Pools,terminal=Terminals; T_PT[pool,terminal]],
					y_P[pool] >= y_PT[pool,terminal]);
@addConstraint(m, Topology_P5[pool=Pools],
					sum{y_SP[source,pool], source=Sources; T_SP[source,pool]} +
					sum{y_PP[pool2,pool], pool2=Pools; T_PP[pool2,pool]} >= y_P[pool]);
@addConstraint(m, Topology_P6[pool=Pools],
					sum{y_PP[pool,pool2], pool2=Pools; T_PP[pool,pool2]} +
					sum{y_PT[pool,terminal], terminal=Terminals; T_PT[pool,terminal]} >= y_P[pool]);

@addConstraint(m, Topology_T1[source=Sources,terminal=Terminals; T_ST[source,terminal]],
					y_T[terminal] >= y_ST[source,terminal]);
@addConstraint(m, Topology_T2[pool=Pools,terminal=Terminals; T_PT[pool,terminal]],
					y_T[terminal] >= y_PT[pool,terminal]);
@addConstraint(m, Topology_T3[terminal=Terminals],
					sum{y_ST[source,terminal], source=Sources; T_ST[source,terminal]} +
					sum{y_PT[pool,terminal], pool=Pools; T_PT[pool,terminal]} >= y_T[terminal]);


# SECOND-STAGE MODELS

@second_stage m scenario begin

    sb = StochasticBlock(m, probability[scenario]);

#----------------------------------------------
#----------	Second-stage variables	-----------
#----------------------------------------------

	@defVar(sb, 0 <= flow_SP[source=Sources,pool=Pools; T_SP[source,pool]] <= 
				flow_SP_up[source,pool]);
	@defVar(sb, 0 <= flow_ST[source=Sources,terminal=Terminals; T_ST[source,terminal]] <= 
				flow_ST_up[source,terminal]);
	@defVar(sb, 0 <= flow_PP[pool=Pools,pool2=Pools,Qualities; T_PP[pool,pool2]] <=
				flow_PP_up[pool,pool2]);
	@defVar(sb, 0 <= flow_PT[pool=Pools,terminal=Terminals,Qualities; T_PT[pool,terminal]] <=
				flow_PT_up[pool,terminal]);
	@defVar(sb, 0 <= fraction_PP[pool=Pools,pool2=Pools; T_PP[pool,pool2]] <= 1);
	@defVar(sb, 0 <= fraction_PT[pool=Pools,terminal=Terminals; T_PT[pool,terminal]] <= 1);
	@defVar(sb, blin1[source=Sources,pool=Pools,pool2=Pools; T_SP[source,pool] && T_PP[pool,pool2]]);
	@defVar(sb, blin2[pool=Pools,pool2=Pools,pool3=Pools,Qualities; T_PP[pool,pool2] && T_PP[pool3,pool]]);
	@defVar(sb, blin3[source=Sources,pool=Pools,terminal=Terminals; T_SP[source,pool] && T_PT[pool,terminal]]);
	@defVar(sb, blin4[pool=Pools,pool2=Pools,terminal=Terminals,Qualities; T_PP[pool2,pool] && T_PP[pool,terminal]]);

#--------------------------------------------------
#----------	Second-stage constraints	-----------
#--------------------------------------------------

# Scenario objectives
    @setObjective(sb, Min,
					sum{Cost_source_flow[source]*(
							sum{flow_SP[source,pool], pool=Pools; T_SP[source,pool]} +
							sum{flow_ST[source,terminal], terminal=Terminals; T_ST[source,terminal]}
							), source=Sources} -
					sum{Price_terminal_flow[terminal]*(
							sum{flow_ST[source,terminal], source=Sources; T_ST[source,terminal]} +
							sum{flow_PT[pool,terminal,quality], pool=Pools, quality=Qualities; T_PT[pool,terminal]}
							), terminal=Terminals} );

# Flow bound constraints
	@addConstraint(sb, Source_flow_up[source=Sources],
					sum{flow_SP[source,pool], pool=Pools; T_SP[source,pool]} +
					sum{flow_ST[source,terminal], terminal=Terminals; T_ST[source,terminal]} <=
					y_S[source]*flow_S_up[source]);
	@addConstraint(sb, Source_flow_lo[source=Sources],
					sum{flow_SP[source,pool], pool=Pools; T_SP[source,pool]} +
					sum{flow_ST[source,terminal], terminal=Terminals; T_ST[source,terminal]} >=
					y_S[source]*flow_S_lo[source]);
	@addConstraint(sb, SP_flow_up[source=Sources,pool=Pools; T_SP[source,pool]],
					flow_SP[source,pool] <= y_SP[source,pool]*flow_SP_up[source,pool]);
	@addConstraint(sb, SP_flow_lo[source=Sources,pool=Pools; T_SP[source,pool]],
					flow_SP[source,pool] >= y_SP[source,pool]*flow_SP_lo[source,pool]);
	@addConstraint(sb, ST_flow_up[source=Sources,terminal=Terminals; T_ST[source,terminal]],
					flow_ST[source,terminal] <= y_ST[source,terminal]*flow_ST_up[source,terminal]);
	@addConstraint(sb, ST_flow_lo[source=Sources,terminal=Terminals; T_ST[source,terminal]],
					flow_ST[source,terminal] >= y_ST[source,terminal]*flow_ST_lo[source,terminal]);
	@addConstraint(sb, PP_flow_up[pool=Pools,pool2=Pools; T_PP[pool,pool2]],
					sum{flow_PP[pool,pool2,quality], quality=Qualities} <= y_PP[pool,pool2]*flow_PP_up[pool,pool2]);
	@addConstraint(sb, PP_flow_lo[pool=Pools,pool2=Pools; T_PP[pool,pool2]],
					sum{flow_PP[pool,pool2,quality], quality=Qualities} >= y_PP[pool,pool2]*flow_PP_lo[pool,pool2]);
	@addConstraint(sb, PT_flow_up[pool=Pools,terminal=Terminals; T_PT[pool,terminal]],
					sum{flow_PT[pool,terminal,quality], quality=Qualities} <= y_PT[pool,terminal]*flow_PT_up[pool,terminal]);
	@addConstraint(sb, PT_flow_lo[pool=Pools,terminal=Terminals; T_PT[pool,terminal]],
					sum{flow_PT[pool,terminal,quality], quality=Qualities} >= y_PT[pool,terminal]*flow_PT_lo[pool,terminal]);

# Flow balances
	@addConstraint(sb, PT_balance[pool=Pools,terminal=Terminals,quality=Qualities; T_PT[pool,terminal]],
					flow_PT[pool,terminal,quality] ==
					sum{Source_quality[source,quality,scenario]*blin3[source,pool,terminal], source=Sources; T_SP[source,pool]} +
					sum{blin4[pool,pool2,terminal,quality], pool2=Pools; T_PP[pool2,pool]});
	@addConstraint(sb, PP_balance[pool=Pools,pool2=Pools,quality=Qualities; T_PP[pool,pool2]],
					flow_PP[pool,pool2,quality] ==
					sum{Source_quality[source,quality,scenario]*blin1[source,pool,pool2], source=Sources; T_SP[source,pool]} +
					sum{blin2[pool,pool2,pool3,quality], pool3=Pools; T_PP[pool3,pool]});
	@addConstraint(sb, Split_balance[pool=Pools],
					sum{fraction_PP[pool,pool2], pool2=Pools; T_PP[pool,pool2]} +
					sum{fraction_PT[pool,terminal], terminal=Terminals; T_PT[pool,terminal]} == 1);

# RLT cuts
	@addConstraint(sb, RLT1[source=Sources,pool=Pools; T_SP[source,pool]],
					flow_SP[source,pool] ==
					sum{blin1[source,pool,pool2], pool2=Pools; T_PP[pool,pool2]} +
					sum{blin3[source,pool,terminal], terminal=Terminals; T_PT[pool,terminal]});
	@addConstraint(sb, RLT2[pool2=Pools,pool=Pools,quality=Qualities; T_PP[pool2,pool]],
					flow_PP[pool2,pool,quality] ==
					sum{blin2[pool,pool3,pool2,quality], pool3=Pools; T_PP[pool,pool3]} +
					sum{blin4[pool,pool2,terminal,quality], terminal=Terminals; T_PT[pool,terminal]});

# Demand constraints
	@addConstraint(sb, Demand_up[terminal=Terminals],
					y_T[terminal]*Terminal_demand_up[terminal,scenario] >=
					sum{flow_ST[source,terminal], source=Sources; T_ST[source,terminal]} +
					sum{flow_PT[pool,terminal,quality], pool=Pools, quality=Qualities; T_PT[pool,terminal]});
	@addConstraint(sb, Demand_lo[terminal=Terminals],
					y_T[terminal]*Terminal_demand_lo[terminal] <=
					sum{flow_ST[source,terminal], source=Sources; T_ST[source,terminal]} +
					sum{flow_PT[pool,terminal,quality], pool=Pools, quality=Qualities; T_PT[pool,terminal]});
	@addConstraint(sb, Terminal_quality_spec_up[terminal=Terminals,quality=Qualities],
					sum{Source_quality[source,quality,scenario]*flow_ST[source,terminal], source=Sources; T_ST[source,terminal]} +
					sum{flow_PT[pool,terminal,quality], pool=Pools; T_PT[pool,terminal]} <=
					Terminal_quality_up[terminal,quality]*(
					sum{flow_ST[source,terminal], source=Sources; T_ST[source,terminal]} +
					sum{flow_PT[pool,terminal,quality2], pool=Pools, quality2=Qualities; T_PT[pool,terminal]}));
	@addConstraint(sb, Terminal_quality_spec_lo[terminal=Terminals,quality=Qualities],
					sum{Source_quality[source,quality,scenario]*flow_ST[source,terminal], source=Sources; T_ST[source,terminal]} +
					sum{flow_PT[pool,terminal,quality], pool=Pools; T_PT[pool,terminal]} >=
					Terminal_quality_lo[terminal,quality]*(
					sum{flow_ST[source,terminal], source=Sources; T_ST[source,terminal]} +
					sum{flow_PT[pool,terminal,quality2], pool=Pools, quality2=Qualities; T_PT[pool,terminal]}));

# Standard McCormick relaxations for the bilinear terms

	@addConstraint(sb, blincon1_1[source=Sources,pool=Pools,pool2=Pools; T_SP[source,pool] && T_PP[pool,pool2]],
					blin1[source,pool,pool2] >= 0);
	@addConstraint(sb, blincon1_2[source=Sources,pool=Pools,pool2=Pools; T_SP[source,pool] && T_PP[pool,pool2]],
					blin1[source,pool,pool2] >= 
					flow_SP_up[source,pool]*fraction_PP[pool,pool2] +
					flow_SP[source,pool]*1 -
					flow_SP_up[source,pool]*1);
	@addConstraint(sb, blincon1_3[source=Sources,pool=Pools,pool2=Pools; T_SP[source,pool] && T_PP[pool,pool2]],
					blin1[source,pool,pool2] <= 
					flow_SP_up[source,pool]*fraction_PP[pool,pool2]);
	@addConstraint(sb, blincon1_4[source=Sources,pool=Pools,pool2=Pools; T_SP[source,pool] && T_PP[pool,pool2]],
					blin1[source,pool,pool2] <= 
					flow_SP[source,pool]*1);

	@addConstraint(sb, blincon2_1[pool=Pools,pool2=Pools,pool3=Pools,quality=Qualities; T_PP[pool,pool2] && T_PP[pool3,pool]],
					blin2[pool,pool2,pool3,quality] >= 0);
	@addConstraint(sb, blincon2_2[pool=Pools,pool2=Pools,pool3=Pools,quality=Qualities; T_PP[pool,pool2] && T_PP[pool3,pool]],
					blin2[pool,pool2,pool3,quality] >= 
					flow_PP_up[pool3,pool]*fraction_PP[pool,pool2] +
					flow_PP[pool3,pool,quality]*1 -
					flow_PP_up[pool3,pool]*1);
	@addConstraint(sb, blincon2_3[pool=Pools,pool2=Pools,pool3=Pools,quality=Qualities; T_PP[pool,pool2] && T_PP[pool3,pool]],
					blin2[pool,pool2,pool3,quality] <= 
					flow_PP_up[pool3,pool]*fraction_PP[pool,pool2]);
	@addConstraint(sb, blincon2_4[pool=Pools,pool2=Pools,pool3=Pools,quality=Qualities; T_PP[pool,pool2] && T_PP[pool3,pool]],
					blin2[pool,pool2,pool3,quality] <= 
					flow_PP[pool3,pool,quality]*1);

	@addConstraint(sb, blincon3_1[source=Sources,pool=Pools,terminal=Terminals; T_SP[source,pool] && T_PT[pool,terminal]],
					blin3[source,pool,terminal] >= 0);
	@addConstraint(sb, blincon3_2[source=Sources,pool=Pools,terminal=Terminals; T_SP[source,pool] && T_PT[pool,terminal]],
					blin3[source,pool,terminal] >= 
					flow_SP_up[source,pool]*fraction_PT[pool,terminal] +
					flow_SP[source,pool]*1 -
					flow_SP_up[source,pool]);
	@addConstraint(sb, blincon3_3[source=Sources,pool=Pools,terminal=Terminals; T_SP[source,pool] && T_PT[pool,terminal]],
					blin3[source,pool,terminal] <= 
					flow_SP_up[source,pool]*fraction_PT[pool,terminal]);
	@addConstraint(sb, blincon3_4[source=Sources,pool=Pools,terminal=Terminals; T_SP[source,pool] && T_PT[pool,terminal]],
					blin3[source,pool,terminal] <= 
					flow_SP[source,pool]*1);

	@addConstraint(sb, blincon4_1[pool=Pools,pool2=Pools,terminal=Terminals,quality=Qualities; T_PT[pool,terminal] && T_PP[pool2,pool]],
					blin4[pool,pool2,terminal,quality] >= 0);
	@addConstraint(sb, blincon4_2[pool=Pools,pool2=Pools,terminal=Terminals,quality=Qualities; T_PT[pool,terminal] && T_PP[pool2,pool]],
					blin4[pool,pool2,terminal,quality] >= 
					flow_PP_up[pool2,pool]*fraction_PT[pool,terminal] +
					flow_PP[pool2,pool,quality]*1 -
					flow_PP_up[pool2,pool]*1);
	@addConstraint(sb, blincon4_3[pool=Pools,pool2=Pools,terminal=Terminals,quality=Qualities; T_PT[pool,terminal] && T_PP[pool2,pool]],
					blin4[pool,pool2,terminal,quality] <= 
					flow_PP_up[pool2,pool]*fraction_PT[pool,terminal]);
	@addConstraint(sb, blincon4_4[pool=Pools,pool2=Pools,terminal=Terminals,quality=Qualities; T_PT[pool,terminal] && T_PP[pool2,pool]],
					blin4[pool,pool2,terminal,quality] <= 
					flow_PP[pool2,pool,quality]*1);

end



